class ChebyshevCurve extends ParametricCurve
{
  // precomputed Chebyshev roots for degree 1, 3, 5, and 7
  float ch1[] = { 0.0, 1.0};
  float ch2[] = { 0.0, 0.5, 1.0};
  float ch3[] = { 0.0, 0.25, 0.75, 1.0};
  float ch4[] = { 0.0, 0.1464466094067262, 0.5, 0.8535533905932737, 1.0};
  float ch5[] = { 0.0, 0.09549150281252627, 0.3454915028125263, 0.6545084971874737, 0.9045084971874737, 1.0};
  float ch6[] = { 0.0, 0.06698729810778065, 0.24999999999999994, 0.49999999999999994, 0.7499999999999999, 0.9330127018922194, 1.0};
  float ch7[] = { 0.0, 0.04951556604879043, 0.1882550990706332, 0.3887395330218428, 0.6112604669781572, 0.8117449009293667, 0.9504844339512095, 1.0};
  float ch8[] = { 0.0, 0.03806023374435663, 0.1464466094067262, 0.3086582838174551, 0.49999999999999994, 0.6913417161825448, 0.8535533905932737, 0.9619397662556434, 1.0};
  
  // uv0, uv1 \in [0,1]^2 correspond to the endpoints of the cell edge (in P1) we are constructing this Hermite interpolation for.
  // This polynomial is constructed by interpolating the mapping values in Chebyshev nodes along the current edge
  ChebyshevCurve( PlanarMapping _mapping, int _n, PVector uv0, PVector uv1 )
  {
    n = 2*_n+1;
    mapping = _mapping;
    
    // sample the mapping at n+1 points, at Chebyshev parameters at that
    PVector pts[] = new PVector[n+1];
    for (int i=0; i<n+1; ++i)
    {
      float t = 0.0;
      switch ( n )
      {
        case 1: t = ch1[i]; break;
        case 2: t = ch2[i]; break;
        case 3: t = ch3[i]; break;
        case 4: t = ch4[i]; break;
        case 5: t = ch5[i]; break;
        case 6: t = ch6[i]; break;
        case 7: t = ch7[i]; break;
        case 8: t = ch8[i]; break;
      }
      float u = lerp( uv0.x, uv1.x, t );
      float v = lerp( uv0.y, uv1.y, t );
      pts[i] = _mapping.f( getPoint( u, v ) );       
    }
    
    // let the Bezier curve construct itself - it will assume Chebyshev node parameters as well
    bez = new BezierN(pts);
    bez.mainColor = color(0,0,0);
    
    // this is a horizontal edge (in the sense of orientation in P1) iff the two UVs are on the same v-isoline
    isHorizontal = ( uv0.y == uv1.y );
    
  }  
  
  PVector eval( float t )
  {
    return eval( new PVector(), t );
  }
  
  PVector eval( PVector pt, float t )
  {
    PVector result = bez.getU(t);
    
    if ( doErrorBound )
    {
      PVector M = mapping.boundDf( n+1 );
      
      // careful: for horizontal lines we have to add the error bound _vertically_ and vice versa!
      M.y *= sx;
      M.x *= sy;
      
      // also careful: we need the absolute of tn, see the phrasing of the remainder and its bound using a derivative bound!
      result = PVector.add( result, PVector.mult( M, 1.0 ) );
    }    
    
    return result;
  }
  
  void draw()
  {
    bez.draw();
  }
  
  PlanarMapping mapping; // the mapping that we approximate
  int n; // degree
  float sx; // which direction to extend the boundary curve
  float sy; // which direction to extend the boundary curve
  PVector p0 = new PVector(), p1 = new PVector();  
  boolean doErrorBound; // whether to use error bounds
  boolean isHorizontal = false;
  BezierN bez;
}
