  class HermiteCurve extends ParametricCurve
{
  // uv0, uv1 \in [0,1]^2 correspond to the endpoints of the cell edge (in P1) we are constructing this Hermite interpolation for.
  // The resulting polynomial will be of degree 2*_n+1
  HermiteCurve( PlanarMapping _mapping, int _n, PVector uv0, PVector uv1 )
  {
    n = 2*_n+1;
    mapping = _mapping;
    
    // the Bezier control points of this Hermite interpolant
    b = new PVector[n+1];
    
    // we use a Bezier curve to draw this - point its control points to the ones we compute here
    bez = new BezierN(n);
    bez.cp = b;
    //bez.toggleCP();
    //bez.togglePolygon();
    bez.mainColor = color(0,0,0);

    // these are our vertices in the P1 plane
    p0 = getPoint(uv0.x, uv0.y);
    p1 = getPoint(uv1.x, uv1.y);

    // the first and last control points are the maps of the boundary edges
    b[0] = mapping.f( p0 );
    b[n] = mapping.f( p1 );
    
    // this is a horizontal edge (in the sense of orientation in P1) iff the two UVs are on the same v-isoline
    isHorizontal = ( uv0.y == uv1.y );
    
    // evaluate the necessary derivatives, if there are any; otherwise: early out
    if ( _n == 0 ) return;
    
    // compute (and store, for convenience) the derivatives at the endpoints 
    m0 = new PVector[_n];
    m1 = new PVector[_n];

    for (int i=0; i<_n; ++i)
    {
      int dx = ( isHorizontal ) ? i+1 : 0;
      int dy = ( isHorizontal ) ? 0   : i+1;
      
      m0[i] = new PVector();
      m1[i] = new PVector();
      m0[i] = mapping.df( dx, dy, p0 );      
      m1[i] = mapping.df( dx, dy, p1 );
    }
    
    if ( _n >= 1 )
    {
      float denom = 1.0/n;
      // complete the next control point at the t=0 endpoint
      b[1] = new PVector();
      b[1].x = m0[0].x * denom + b[0].x; 
      b[1].y = m0[0].y * denom + b[0].y; 
      
      // and do the same for the t=1 endpoint
      b[n-1] = new PVector();
      b[n-1].x = b[n].x - m1[0].x * denom; 
      b[n-1].y = b[n].y - m1[0].y * denom; 
    }
    
    if ( _n >= 2 )
    {
      float denom = 1.0 / ( n*(n-1) );
      // complete the next control point at the t=0 endpoint
      b[2] = new PVector();
      b[2].x = m0[1].x * denom + ( b[1].x - b[0].x ) +  b[1].x; 
      b[2].y = m0[1].y * denom + ( b[1].y - b[0].y ) +  b[1].y; 
      
      // and do the same for the t=1 endpoint
      b[n-2] = new PVector();
      b[n-2].x = b[n-1].x - ( b[n].x - b[n-1].x ) + m1[1].x * denom; 
      b[n-2].y = b[n-1].y - ( b[n].y - b[n-1].y ) + m1[1].y * denom; 
    }
    
    if ( _n >= 3 )
    {
      float denom = 1.0 / ( n*(n-1)*(n-2) );
      // complete the next control point at the t=0 endpoint
      b[3] = new PVector();
      b[3].x = m0[2].x / denom + ( b[2].x - 2*b[1].x + b[0].x ) + ( b[2].x - b[1].x ) +  b[2].x; 
      b[3].y = m0[2].y / denom + ( b[2].y - 2*b[1].y + b[0].y )  + ( b[2].y - b[1].y ) +  b[2].y; 
      
      // and do the same for the t=1 endpoint
      b[n-3] = new PVector();
      b[n-3].x = b[n-2].x - ( b[n-1].x - b[n-2].x ) + ( b[n].x - 2*b[n-1].x + b[n-2].x ) - m1[2].x * denom; 
      b[n-3].y = b[n-2].y - ( b[n-1].y - b[n-2].y ) + ( b[n].y - 2*b[n-1].y + b[n-2].y ) - m1[2].y * denom; 
    }
  }  
  
  PVector eval( float t )
  {
    return eval( new PVector(), t );
  }
  
  PVector eval( PVector pt, float t )
  {
    PVector result = bez.getU(t);
    
    if ( doErrorBound )
    {
      PVector M = mapping.boundDf( n+1 );
      
      // careful: for horizontal lines we have to add the error bound _vertically_ and vice versa!
      M.y *= sx;
      M.x *= sy;
      
      // also careful: we need the absolute of tn, see the phrasing of the remainder and its bound using a derivative bound!
      result = PVector.add( result, PVector.mult( M, 1.0 ) );
    }    
    
    return result;
  }
  
  void draw()
  {
    bez.draw();
  }
  
  PlanarMapping mapping; // the mapping that we approximate
  int n; // degree
  float sx; // which direction to extend the boundary curve
  float sy; // which direction to extend the boundary curve
  PVector b[]; // the control points (in P2)
  PVector p0 = new PVector(), p1 = new PVector();
  PVector m0[], m1[]; // array of derivatives at the boundary edge vertices  
  boolean doErrorBound; // whether to use error bounds
  boolean isHorizontal = false;
  BezierN bez;
}
